"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BinarySensorCCBehaviors = void 0;
const cc_1 = require("@zwave-js/cc");
const core_1 = require("@zwave-js/core");
const testing_1 = require("@zwave-js/testing");
const defaultCapabilities = {
    supportedSensorTypes: [],
};
const respondToBinarySensorSupportedGet = {
    async onControllerFrame(controller, self, frame) {
        if (frame.type === testing_1.MockZWaveFrameType.Request
            && frame.payload instanceof cc_1.BinarySensorCCSupportedGet) {
            const capabilities = {
                ...defaultCapabilities,
                ...self.getCCCapabilities(core_1.CommandClasses["Binary Sensor"], frame.payload.endpointIndex),
            };
            const cc = new cc_1.BinarySensorCCSupportedReport(self.host, {
                nodeId: controller.host.ownNodeId,
                supportedSensorTypes: capabilities.supportedSensorTypes,
            });
            await self.sendToController((0, testing_1.createMockZWaveRequestFrame)(cc, {
                ackRequested: false,
            }));
            return true;
        }
        return false;
    },
};
const respondToBinarySensorGet = {
    async onControllerFrame(controller, self, frame) {
        if (frame.type === testing_1.MockZWaveFrameType.Request
            && frame.payload instanceof cc_1.BinarySensorCCGet) {
            const capabilities = {
                ...defaultCapabilities,
                ...self.getCCCapabilities(core_1.CommandClasses["Binary Sensor"], frame.payload.endpointIndex),
            };
            let sensorType;
            if (frame.payload.sensorType == undefined
                || frame.payload.sensorType === cc_1.BinarySensorType.Any) {
                // If the sensor type is not specified, use the first supported one
                sensorType = capabilities.supportedSensorTypes[0];
            }
            else {
                sensorType = frame.payload.sensorType;
            }
            if (sensorType != undefined) {
                const value = capabilities.getValue?.(sensorType) ?? false;
                const cc = new cc_1.BinarySensorCCReport(self.host, {
                    nodeId: controller.host.ownNodeId,
                    type: sensorType,
                    value,
                });
                await self.sendToController((0, testing_1.createMockZWaveRequestFrame)(cc, {
                    ackRequested: false,
                }));
            }
            return true;
        }
        return false;
    },
};
exports.BinarySensorCCBehaviors = [
    respondToBinarySensorSupportedGet,
    respondToBinarySensorGet,
];
//# sourceMappingURL=BinarySensor.js.map