"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ZnifferCCParsingContext = void 0;
/* eslint-disable @typescript-eslint/no-unused-vars */
const cc_1 = require("@zwave-js/cc");
const config_1 = require("@zwave-js/config");
const core_1 = require("@zwave-js/core");
class ZnifferCCParsingContext {
    ownNodeId;
    homeId;
    securityManager;
    securityManager2;
    securityManagerLR;
    constructor(ownNodeId, homeId, securityManager, securityManager2, securityManagerLR) {
        this.ownNodeId = ownNodeId;
        this.homeId = homeId;
        this.securityManager = securityManager;
        this.securityManager2 = securityManager2;
        this.securityManagerLR = securityManagerLR;
    }
    getSafeCCVersion(cc, nodeId, endpointIndex) {
        // We don't know any versions of the node. Try parsing with the highest version we support
        return (0, cc_1.getImplementedVersion)(cc);
    }
    getSupportedCCVersion(cc, nodeId, endpointIndex) {
        // We don't know any versions of the node. Try parsing with the highest version we support
        return (0, cc_1.getImplementedVersion)(cc);
    }
    isCCSecure(cc, nodeId, endpointIndex) {
        // Don't care when parsing
        return false;
    }
    getHighestSecurityClass(nodeId) {
        return core_1.SecurityClass.S2_AccessControl;
    }
    hasSecurityClass(nodeId, securityClass) {
        // We don't actually know. Attempt parsing with all security classes
        return true;
    }
    setSecurityClass(nodeId, securityClass, granted) {
        // Do nothing
    }
    getNextCallbackId() {
        throw new Error("Method not implemented.");
    }
    getNextSupervisionSessionId(nodeId) {
        throw new Error("Method not implemented.");
    }
    getDeviceConfig(nodeId) {
        // Disable strict validation while parsing certain CCs
        // Most of this stuff isn't actually needed, only the compat flags...
        return new config_1.DeviceConfig("unknown.json", false, "UNKNOWN_MANUFACTURER", 0x0000, "UNKNOWN_PRODUCT", "UNKNOWN_DESCRIPTION", [], {
            min: "0.0",
            max: "255.255",
        }, true, undefined, undefined, undefined, undefined, 
        // ...down here:
        {
            disableStrictEntryControlDataValidation: true,
            disableStrictMeasurementValidation: true,
        });
    }
}
exports.ZnifferCCParsingContext = ZnifferCCParsingContext;
//# sourceMappingURL=CCParsingContext.js.map