"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.LRNodeInfoFileV5 = exports.nodeIdToLRNodeInfoFileIDV5 = exports.LRNodeInfoFileV5IDBase = exports.NodeInfoFileV1 = exports.nodeIdToNodeInfoFileIDV1 = exports.NodeInfoFileV1IDBase = exports.NodeInfoFileV0 = exports.nodeIdToNodeInfoFileIDV0 = exports.NodeInfoFileV0IDBase = exports.LR_NODEINFOS_PER_FILE_V5 = exports.NODEINFOS_PER_FILE_V1 = void 0;
const safe_1 = require("@zwave-js/core/safe");
const safe_2 = require("@zwave-js/shared/safe");
const NVMFile_1 = require("./NVMFile");
exports.NODEINFOS_PER_FILE_V1 = 4;
exports.LR_NODEINFOS_PER_FILE_V5 = 50;
const NODEINFO_SIZE = 1 + 5 + safe_1.NUM_NODEMASK_BYTES;
const LR_NODEINFO_SIZE = 3;
const EMPTY_NODEINFO_FILL = 0xff;
const emptyNodeInfo = Buffer.alloc(NODEINFO_SIZE, EMPTY_NODEINFO_FILL);
function parseNodeInfo(nodeId, buffer, offset) {
    const { hasSpecificDeviceClass, ...protocolInfo } = (0, safe_1.parseNodeProtocolInfo)(buffer, offset);
    const genericDeviceClass = buffer[offset + 3];
    const specificDeviceClass = hasSpecificDeviceClass
        ? buffer[offset + 4]
        : null;
    const neighbors = (0, safe_1.parseBitMask)(buffer.subarray(offset + 5, offset + 5 + safe_1.NUM_NODEMASK_BYTES));
    const sucUpdateIndex = buffer[offset + 5 + safe_1.NUM_NODEMASK_BYTES];
    return {
        nodeId,
        ...protocolInfo,
        genericDeviceClass,
        specificDeviceClass,
        neighbors,
        sucUpdateIndex,
    };
}
function encodeNodeInfo(nodeInfo) {
    const ret = Buffer.alloc(NODEINFO_SIZE);
    const hasSpecificDeviceClass = nodeInfo.specificDeviceClass != null;
    const protocolInfo = {
        ...(0, safe_2.pick)(nodeInfo, [
            "isListening",
            "isFrequentListening",
            "isRouting",
            "supportedDataRates",
            "protocolVersion",
            "optionalFunctionality",
            "nodeType",
            "supportsSecurity",
            "supportsBeaming",
        ]),
        hasSpecificDeviceClass,
    };
    (0, safe_1.encodeNodeProtocolInfo)(protocolInfo).copy(ret, 0);
    ret[3] = nodeInfo.genericDeviceClass;
    if (hasSpecificDeviceClass)
        ret[4] = nodeInfo.specificDeviceClass;
    (0, safe_1.encodeBitMask)(nodeInfo.neighbors, safe_1.MAX_NODES).copy(ret, 5);
    ret[5 + safe_1.NUM_NODEMASK_BYTES] = nodeInfo.sucUpdateIndex;
    return ret;
}
function parseLRNodeInfo(nodeId, buffer, offset) {
    // The node info in LR NVM files is packed:
    // Byte 0 CAPABILITY:
    //   Bit 0: Routing (?)
    //   Bit 1: Listening
    //   Bit 2: has specific device class (?)
    //   Bit 3: Beam capability
    //   Bit 4: Optional functionality
    //   Bits 5-6: FLiRS
    //   Bit 7: Unused (?)
    // Byte 1: Generic device class
    // Byte 2: Specific device class
    // Protocol version is always 3
    // Security is always true
    // Supported speed is always 100kbps (speed = 0, speed ext = 2)
    // Never: routing end node, controller
    const capability = buffer[offset];
    const isRouting = !!(capability & 0b0000_0001); // ZWLR Mesh??
    const isListening = !!(capability & 0b0000_0010);
    const hasSpecificDeviceClass = !!(capability & 0b0000_0100);
    const supportsBeaming = !!(capability & 0b0000_1000);
    const optionalFunctionality = !!(capability & 0b0001_0000);
    let isFrequentListening;
    switch (capability & 0b0110_0000) {
        case 0b0100_0000:
            isFrequentListening = "1000ms";
            break;
        case 0b0010_0000:
            isFrequentListening = "250ms";
            break;
        default:
            isFrequentListening = false;
    }
    const nodeType = safe_1.NodeType["End Node"];
    const supportsSecurity = true;
    const protocolVersion = 3;
    const supportedDataRates = [100000];
    return {
        nodeId,
        isRouting,
        isListening,
        supportsBeaming,
        isFrequentListening,
        optionalFunctionality,
        nodeType,
        supportsSecurity,
        protocolVersion,
        supportedDataRates,
        genericDeviceClass: buffer[offset + 1],
        specificDeviceClass: hasSpecificDeviceClass ? buffer[offset + 2] : null,
    };
}
function encodeLRNodeInfo(nodeInfo) {
    const ret = Buffer.alloc(LR_NODEINFO_SIZE);
    let capability = 0;
    if (nodeInfo.isRouting)
        capability |= 0b0000_0001;
    if (nodeInfo.isListening)
        capability |= 0b0000_0010;
    if (nodeInfo.specificDeviceClass != null)
        capability |= 0b0000_0100;
    if (nodeInfo.supportsBeaming)
        capability |= 0b0000_1000;
    if (nodeInfo.optionalFunctionality)
        capability |= 0b0001_0000;
    if (nodeInfo.isFrequentListening === "1000ms") {
        capability |= 0b0100_0000;
    }
    else if (nodeInfo.isFrequentListening === "250ms") {
        capability |= 0b0010_0000;
    }
    ret[0] = capability;
    ret[1] = nodeInfo.genericDeviceClass;
    ret[2] = nodeInfo.specificDeviceClass ?? 0;
    return ret;
}
exports.NodeInfoFileV0IDBase = 0x50100;
function nodeIdToNodeInfoFileIDV0(nodeId) {
    return exports.NodeInfoFileV0IDBase + nodeId - 1;
}
exports.nodeIdToNodeInfoFileIDV0 = nodeIdToNodeInfoFileIDV0;
let NodeInfoFileV0 = class NodeInfoFileV0 extends NVMFile_1.NVMFile {
    constructor(options) {
        super(options);
        if ((0, NVMFile_1.gotDeserializationOptions)(options)) {
            this.nodeInfo = parseNodeInfo(this.fileId - exports.NodeInfoFileV0IDBase + 1, this.payload, 0);
        }
        else {
            this.nodeInfo = options.nodeInfo;
        }
    }
    nodeInfo;
    serialize() {
        this.fileId = nodeIdToNodeInfoFileIDV0(this.nodeInfo.nodeId);
        this.payload = encodeNodeInfo(this.nodeInfo);
        return super.serialize();
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    toJSON() {
        return {
            ...super.toJSON(),
            nodeInfo: this.nodeInfo,
        };
    }
};
exports.NodeInfoFileV0 = NodeInfoFileV0;
exports.NodeInfoFileV0 = NodeInfoFileV0 = __decorate([
    (0, NVMFile_1.nvmFileID)((id) => id >= exports.NodeInfoFileV0IDBase && id < exports.NodeInfoFileV0IDBase + safe_1.MAX_NODES)
], NodeInfoFileV0);
exports.NodeInfoFileV1IDBase = 0x50200;
function nodeIdToNodeInfoFileIDV1(nodeId) {
    return (exports.NodeInfoFileV1IDBase + Math.floor((nodeId - 1) / exports.NODEINFOS_PER_FILE_V1));
}
exports.nodeIdToNodeInfoFileIDV1 = nodeIdToNodeInfoFileIDV1;
let NodeInfoFileV1 = class NodeInfoFileV1 extends NVMFile_1.NVMFile {
    constructor(options) {
        super(options);
        if ((0, NVMFile_1.gotDeserializationOptions)(options)) {
            this.nodeInfos = [];
            for (let i = 0; i < exports.NODEINFOS_PER_FILE_V1; i++) {
                const nodeId = (this.fileId - exports.NodeInfoFileV1IDBase)
                    * exports.NODEINFOS_PER_FILE_V1
                    + 1
                    + i;
                const offset = i * NODEINFO_SIZE;
                const entry = this.payload.subarray(offset, offset + NODEINFO_SIZE);
                if (entry.equals(emptyNodeInfo))
                    continue;
                const nodeInfo = parseNodeInfo(nodeId, entry, 0);
                this.nodeInfos.push(nodeInfo);
            }
        }
        else {
            this.nodeInfos = options.nodeInfos;
        }
    }
    nodeInfos;
    serialize() {
        // The infos must be sorted by node ID
        this.nodeInfos.sort((a, b) => a.nodeId - b.nodeId);
        const minNodeId = this.nodeInfos[0].nodeId;
        this.fileId = nodeIdToNodeInfoFileIDV1(minNodeId);
        this.payload = Buffer.alloc(NODEINFO_SIZE * exports.NODEINFOS_PER_FILE_V1, EMPTY_NODEINFO_FILL);
        const minFileNodeId = Math.floor((minNodeId - 1) / exports.NODEINFOS_PER_FILE_V1)
            * exports.NODEINFOS_PER_FILE_V1
            + 1;
        for (const nodeInfo of this.nodeInfos) {
            const offset = (nodeInfo.nodeId - minFileNodeId) * NODEINFO_SIZE;
            encodeNodeInfo(nodeInfo).copy(this.payload, offset);
        }
        return super.serialize();
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    toJSON() {
        return {
            ...super.toJSON(),
            "node infos": this.nodeInfos,
        };
    }
};
exports.NodeInfoFileV1 = NodeInfoFileV1;
exports.NodeInfoFileV1 = NodeInfoFileV1 = __decorate([
    (0, NVMFile_1.nvmFileID)((id) => id >= exports.NodeInfoFileV1IDBase
        && id < exports.NodeInfoFileV1IDBase + safe_1.MAX_NODES / exports.NODEINFOS_PER_FILE_V1)
], NodeInfoFileV1);
exports.LRNodeInfoFileV5IDBase = 0x50800;
function nodeIdToLRNodeInfoFileIDV5(nodeId) {
    return (exports.LRNodeInfoFileV5IDBase
        + Math.floor((nodeId - 256) / exports.LR_NODEINFOS_PER_FILE_V5));
}
exports.nodeIdToLRNodeInfoFileIDV5 = nodeIdToLRNodeInfoFileIDV5;
// Counting starts with 5, because we only implemented this after reaching protocol file format 5
let LRNodeInfoFileV5 = class LRNodeInfoFileV5 extends NVMFile_1.NVMFile {
    constructor(options) {
        super(options);
        if ((0, NVMFile_1.gotDeserializationOptions)(options)) {
            this.nodeInfos = [];
            for (let i = 0; i < exports.LR_NODEINFOS_PER_FILE_V5; i++) {
                const nodeId = (this.fileId - exports.LRNodeInfoFileV5IDBase)
                    * exports.LR_NODEINFOS_PER_FILE_V5
                    + 256
                    + i;
                const offset = i * LR_NODEINFO_SIZE;
                const entry = this.payload.subarray(offset, offset + LR_NODEINFO_SIZE);
                if (entry.equals(emptyNodeInfo))
                    continue;
                const nodeInfo = parseLRNodeInfo(nodeId, entry, 0);
                this.nodeInfos.push(nodeInfo);
            }
        }
        else {
            this.nodeInfos = options.nodeInfos;
        }
    }
    nodeInfos;
    serialize() {
        // The infos must be sorted by node ID
        this.nodeInfos.sort((a, b) => a.nodeId - b.nodeId);
        const minNodeId = this.nodeInfos[0].nodeId;
        this.fileId = nodeIdToLRNodeInfoFileIDV5(minNodeId);
        this.payload = Buffer.alloc(LR_NODEINFO_SIZE * exports.LR_NODEINFOS_PER_FILE_V5, EMPTY_NODEINFO_FILL);
        const minFileNodeId = Math.floor((minNodeId - 256) / exports.LR_NODEINFOS_PER_FILE_V5)
            * exports.LR_NODEINFOS_PER_FILE_V5
            + 256;
        for (const nodeInfo of this.nodeInfos) {
            const offset = (nodeInfo.nodeId - minFileNodeId) * LR_NODEINFO_SIZE;
            encodeLRNodeInfo(nodeInfo).copy(this.payload, offset);
        }
        return super.serialize();
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    toJSON() {
        return {
            ...super.toJSON(),
            "node infos": this.nodeInfos,
        };
    }
};
exports.LRNodeInfoFileV5 = LRNodeInfoFileV5;
exports.LRNodeInfoFileV5 = LRNodeInfoFileV5 = __decorate([
    (0, NVMFile_1.nvmFileID)((id) => id >= exports.LRNodeInfoFileV5IDBase
        && id
            < exports.LRNodeInfoFileV5IDBase + safe_1.MAX_NODES_LR / exports.LR_NODEINFOS_PER_FILE_V5)
], LRNodeInfoFileV5);
//# sourceMappingURL=NodeInfoFiles.js.map